/*
 * (C) Daniel M. Duley (mosfet@kde.org) KDE artistic license.
 */
#include "pagerapplet.h"
#include <qpainter.h>
#include <qdrawutil.h>
#include <qtooltip.h>
#include <kconfig.h>
#include <kwin.h>
#include <kwinmodule.h>
#include <kapp.h>
#include <kglobal.h>
#include <klocale.h>
#include <dcopclient.h>

KWinModule *kwin_module;

extern "C"
{
    InternalApplet* init(QWidget *parent, KWinModule *module)
    {
        kwin_module = module;
        return new KMiniPager(parent, "kminipagerapplet");
    }
}

void KMiniPagerArrow::paintEvent(QPaintEvent *)
{
    QPainter p(this);
    qDrawArrow(&p, arrow, Qt::WindowsStyle, isOn() || isDown(), 0, 0, width(),
               height(), colorGroup(), true);
    qDrawWinPanel(&p, 0, 0, width(), height(), colorGroup(), isOn() || isDown());
}

void KMiniPagerArrow::setPosition(InternalApplet::Position p)
{
    switch(p){
    case InternalApplet::Left:
        arrow = Qt::RightArrow; break;
    case InternalApplet::Right:
        arrow = Qt::LeftArrow; break;
    case InternalApplet::Top:
        arrow = Qt::DownArrow; break;
    default:
        arrow = Qt::UpArrow; break;
    }
}

KMiniPagerButton::KMiniPagerButton(int desk, QWidget *parent, const char *name)
    : QButton(parent, name)
{
    deskStr = deskStr.number(desk);
    deskNum = desk;
    connect(this, SIGNAL(clicked()), SLOT(slotClicked()));
}

void KMiniPagerButton::paintEvent(QPaintEvent *)
{
    QPainter p(this);
    p.setFont(font());
    if(!isOn() && !isDown()){
        p.fillRect(rect(), colorGroup().brush(QColorGroup::Dark));
        p.setPen(colorGroup().light());
        p.drawLine(0, 0, width()-1, 0);
        p.drawLine(0, 0, 0, height()-1);
        p.setPen(Qt::black);
        p.drawLine(0, height()-1, width()-1, height()-1);
        p.drawLine(width()-1, 0, width()-1, height()-1);
        p.setPen(Qt::white);
    }
    else{
        p.fillRect(rect(), colorGroup().brush(QColorGroup::Base));
        p.setPen(colorGroup().text());
        p.drawRect(rect());
    }
    p.drawText(0, 0, width(), height(), AlignCenter, deskStr);
}

void KMiniPagerButton::slotClicked()
{
    emit deskChange(deskNum);
}

KMiniPager::KMiniPager(QWidget *parent, const char *name)
    : InternalApplet(parent, name)
{
    launchBtn = new KMiniPagerArrow(this);
    QToolTip::add(launchBtn, i18n("Extended Pager"));
    curDesk = KWin::currentDesktop();
    mainlayout = NULL;
    allocateButtons();
    connect(kwin_module, SIGNAL(desktopChange(int)), SLOT(slotSetDesktop(int)));
    connect(kwin_module, SIGNAL(desktopNumberChange(int)),
            SLOT(slotSetDesktopCount(int)));
    KConfig *config = KGlobal::config();
    config->setGroup("minipager");
    QFont defFont("Helvetica", 10, QFont::Bold);
    defFont = config->readFontEntry("Font", &defFont);
    setFontPropagation(AllChildren);
    setFont(defFont);
}

void KMiniPager::slotSetDesktop(int desktop)
{
    if(KWin::numberOfDesktops() > btnList.count())
        slotSetDesktopCount(desktop);
    btnList[curDesk-1]->setDown(false);
    curDesk = desktop;
    btnList[curDesk-1]->setDown(true);
}

void KMiniPager::slotButtonClicked(int desk)
{
    if(desk == curDesk) // clicked the current desktop
        btnList[curDesk-1]->setDown(true);
    else
        KWin::setCurrentDesktop( desk );
    }

void KMiniPager::setupGeometry(Orientation orient, int w, int h)
{
    InternalApplet::setupGeometry(orient, w, h);
    if(mainlayout)
        delete mainlayout;
    int deskNum = KWin::numberOfDesktops();
    bool horiz = orient == Horizontal;
    Position location = position();
    launchBtn->setPosition(location);
    if(horiz && h <= 32 || !horiz && h <= 32){
        setupSmallGeometry(orient, w, h, deskNum, location);
        return;
    }

    KMiniPagerButton *btn;
    int deskHalf = deskNum/2;
    if(horiz){
        setMinimumSize(deskHalf*((h-10)/2), 32);
        resize(deskHalf*((h-10)/2), h);
        launchBtn->setMinimumSize(1, 10);
        layout = new QGridLayout(2, deskHalf, 0);
        mainlayout = new QVBoxLayout(this, 0);
    }
    else{
        setMinimumSize(32, deskHalf*((w-10)/2));
        resize(w, deskHalf*((w-10)/2));
        launchBtn->setMinimumSize(10, 1);
        layout = new QGridLayout(deskHalf, 2, 0);
        mainlayout = new QHBoxLayout(this, 0);
    }
    if(location == Bottom || location == Right){
        mainlayout->addWidget(launchBtn);
        mainlayout->addLayout(layout, 1);
    }
    else{
        mainlayout->addLayout(layout, 1);
        mainlayout->addWidget(launchBtn);
    }
    int c = 0, i=0;
    while(c <  deskHalf){
        btn = btnList[i];
        if(horiz)
            layout->addWidget(btn, 0, c);
        else
            layout->addWidget(btn, c, 0);
        ++i;
        btn = btnList[i];
        if(horiz)
            layout->addWidget(btn, 1, c);
        else
            layout->addWidget(btn, c, 1);
        ++c, ++i;
    }
    mainlayout->activate();
    if(KWin::numberOfDesktops() > 1){
        if(!isVisible())
            show();
    }
    else
        hide();
}

void KMiniPager::setupSmallGeometry(Orientation orient, int w, int h,
                                    int deskNum, Position location)
{
    bool horiz = orient == Horizontal;
    if(horiz){
        setMinimumSize(deskNum*(h-10), 1);
        resize(deskNum*(h-10), h);
        launchBtn->setMinimumSize(1, 10);
        layout = new QGridLayout(1, deskNum, 0);
        mainlayout = new QVBoxLayout(this, 0);
    }
    else{
        setMinimumSize(1, deskNum*(w-10));
        resize(w, deskNum*(w-10));
        launchBtn->setMinimumSize(10, 1);
        layout = new QGridLayout(deskNum, 1, 0);
        mainlayout = new QHBoxLayout(this, 0);
    }
    if(location == Bottom || location == Right){
        mainlayout->addWidget(launchBtn);
        mainlayout->addLayout(layout, 1);
    }
    else{
        mainlayout->addLayout(layout, 1);
        mainlayout->addWidget(launchBtn);
    }
    int c = 0;
    KMiniPagerButton *btn;
    while(c <  deskNum){
        btn = btnList[c];
        if(horiz)
            layout->addWidget(btn, 0, c);
        else
            layout->addWidget(btn, c, 0);
        ++c;
    }
    mainlayout->activate();
    if(KWin::numberOfDesktops() > 1){
        if(!isVisible())
            show();
    }
    else
        hide();
}


void KMiniPager::allocateButtons()
{
    int i;
    int deskNum = KWin::numberOfDesktops();
    int act = KWin::currentDesktop();
    KMiniPagerButton *btn;
    for(i=1; i <= deskNum; ++i){
        btn = new KMiniPagerButton(i, this);
        btn->setDown(i == act);
        // fix this once kwin supports names
        //QToolTip::add(btn, i18n("Desktop: ") + KWin::desktopName(i));
        QToolTip::add(btn, i18n("Desktop: ") + QString::number(i));

        btnList.append(btn);
        connect(btn, SIGNAL(deskChange(int)), SLOT(slotButtonClicked(int)));
    }
}

void KMiniPager::slotSetDesktopCount(int)
{
    QValueList<KMiniPagerButton*>::Iterator it;
    for(it=btnList.begin(); it != btnList.end(); ++it)
        delete (*it);
    btnList.clear();
    allocateButtons();
    curDesk = KWin::currentDesktop();
    setupGeometry(orientation(), width(), height());
}

#include "pagerapplet.moc"
