#include <qtooltip.h>
#include <qdrawutil.h>
#include <qlayout.h>
#include <kwin.h>
#include <kwm.h>
#include <kwinmodule.h>
#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdrawutil.h>
#include <qpopupmenu.h>

#include "taskbar.h"


KWinModule* kwin_module;

extern "C"
{
    InternalApplet* init(QWidget *parent, KWinModule *module)
    {
        kwin_module = module;
        return new TaskbarApplet(parent, "ktaskbarapplet");

    }
}

TaskbarApplet::TaskbarApplet(QWidget* parent, const char* name )
    : InternalApplet(parent, name )
{
    connect( kwin_module, SIGNAL( windowAdd(WId) ), this, SLOT( windowAdd(WId) ) );
    connect( kwin_module, SIGNAL( windowRemove(WId) ), this, SLOT( windowRemove(WId) ) );
    connect( kwin_module, SIGNAL( windowChange(WId) ), this, SLOT( windowChange(WId) ) );
    connect( kwin_module, SIGNAL( windowActivate(WId) ), this, SLOT( windowActivate(WId) ) );
    connect( kwin_module, SIGNAL( desktopChange(int) ), this, SLOT( desktopChange(int) ) );
    setStretch(Huge);
    updateLayout();
}

TaskbarApplet::~TaskbarApplet()
{
}


TaskButton* TaskbarApplet::findButton( WId w )
{
    for ( TaskButton* b = buttons.first(); b; b = buttons.next() ) {
	if ( b->window() == w )
	    return b;
    }
    return 0;
}

void TaskbarApplet::windowAdd(WId w )
{
    if ( w == topLevelWidget()->winId() )
	return;
    TaskButton* b = new TaskButton(w, this);
    buttons.append( b );
    updateLayout();
}

void TaskbarApplet::windowRemove(WId w )
{
    if ( w == topLevelWidget()->winId() )
	return;
    TaskButton* b = findButton( w );
    if ( b ) {
	buttons.removeRef(b);
	delete b;
	updateLayout();
    }
}

void TaskbarApplet::windowChange(WId w)
{
    TaskButton* b = findButton( w );
    if ( !b )
	return;
    
    b->refresh();
    if ( b->onCurrentDesktop() != isVisible() ) {
	b->isVisible()?b->hide():b->show();
	updateLayout();
    }
    
}

void TaskbarApplet::windowActivate(WId w)
{
    if ( w == topLevelWidget()->winId() )
	return;
    TaskButton* b = findButton( w );
    if ( b )
	b->setActiveButton();
    else
	TaskButton::noActiveButton();
}

void TaskbarApplet::desktopChange(int d)
{
    updateLayout();
}

void TaskbarApplet::setupGeometry(Orientation orient, int appw, int apph)
{
    InternalApplet::setupGeometry(orient, appw, apph);
    lastH = apph;
    updateLayout();
}

void TaskbarApplet::updateLayout()
{
     bool single = false;
    static QGridLayout *layout = NULL;
    if(layout)
        delete layout;
    bool horiz = orientation() == Horizontal;
    int items, i, row;

    TaskButton *b;
    for(items=0, b = buttons.first(); b; b = buttons.next())
        items += b->onCurrentDesktop()?1:0;

    if(items){
        if(horiz && lastH > 32){
            if(items > 1) {
                int offset = items & 1;
                items /= 2;
                items += offset;
            }
            else if(items == 1)
                single = true;
        }
        layout = new QGridLayout(this);
    }
    else
        layout = NULL;
    for(i=0, row=0, b = buttons.first(); b; b = buttons.next()){
        if(b->onCurrentDesktop()){
            b->show();
            if(horiz){
                b->setMinimumSize(64, 8);
                b->setMaximumSize(200, 100);
                layout->addWidget(b, row, i);
                layout->setColStretch(i, 100);
            }
            else{
                b->setMinimumSize(8, 21);
                layout->addWidget(b, i, 0);
            }
            ++i;
            if(i >= items){
                ++row;
                i = 0;
            }
        }
        else{
            b->move(0, 0);
            b->hide();
        }
    }
    if(layout){
        if(!horiz){ // hack since were using the taskbar as required stretch
            layout->setRowStretch(items, 100);
        }
        else{
            if(single)
                layout->addRowSpacing(1, 1);
            layout->setColStretch(items, 1);
        }
        show();
    }
    else
        hide();
}


TaskButton* TaskButton::activeButton = 0;

void TaskButton::setActiveButton()
{
    setOn( TRUE );
    if ( activeButton && activeButton != this )
	activeButton->setOn( FALSE );
    setOn( TRUE );
    activeButton = this;
}

void TaskButton::noActiveButton()
{
    if ( activeButton )
	activeButton->setOn( FALSE );
    activeButton = 0;
}

TaskButton::TaskButton( WId w, QWidget * parent, const char *name )
    : QPushButton( parent, name )
{
    setToggleButton( TRUE );
    win = w;

    refresh();
    connect( this, SIGNAL( clicked() ), this, SLOT( toggled() ) );

    popup = 0;
    desk_popup = 0;
}

TaskButton::~TaskButton()
{
    if ( activeButton == this )
	activeButton = 0;
    delete popup; popup = 0;
}

bool TaskButton::onCurrentDesktop() const
{
    return sticky || kwin_module->currentDesktop() == desktop;
}

void TaskButton::refresh()
{
    pixmap = KWM::miniIcon( win, 16, 16 );
    if(pixmap.isNull())
        pixmap = KGlobal::iconLoader()->loadIcon("bx2", KIconLoader::Small);
    QString t = KWM::titleWithState( win );
    setText( t );
    QToolTip::add( this, t );
    desktop = KWM::desktop( win );
    sticky = KWM::isSticky( win );
}

void TaskButton::drawButton(QPainter *p)
{
    bool sunken = isOn() || isDown();
    kDrawNextButton(p, rect(), colorGroup(), sunken, sunken ?
                    &colorGroup().brush(QColorGroup::Mid) :
                    &colorGroup().brush(QColorGroup::Button));

    const int PIXMAP_AREA_WIDTH = 20;
    QRect br( style().buttonRect( 0, 0, width(), height() ) );

    if ( !pixmap.isNull() ) {
        int dx = ( PIXMAP_AREA_WIDTH - pixmap.width() ) / 2;
        int dy = ( height() - pixmap.height() ) / 2;
        if ( isDown() ) {
            dx++;
            dy++;
        }
        p->drawPixmap( br.x()+dx, dy, pixmap );
    }
    if(isOn() || isDown())
        p->setPen(colorGroup().light());
    else
        p->setPen(colorGroup().buttonText());

    if (!text().isNull()){
        QString s2 = text();
        if (fontMetrics().width(s2) > br.width()-PIXMAP_AREA_WIDTH){
            while (s2.length()>0 &&
                   fontMetrics().width(s2) > br.width() - PIXMAP_AREA_WIDTH
                   - fontMetrics().width("...")) {
                s2.truncate( s2.length() - 1 );
            }
            s2.append("...");
        }
        p->drawText(br.x()+ PIXMAP_AREA_WIDTH, 0, width()-PIXMAP_AREA_WIDTH, height(), AlignLeft|AlignVCenter, s2);
    }
}

void TaskButton::toggled()
{
    if ( isOn() ) {
	setActiveButton();
	KWin::setActiveWindow( win );
    }
    else
	KWM::setIconify( win, TRUE );
}

void TaskButton::clientPopupAboutToShow()
{
    if ( !popup )
        return;
    popup->setItemChecked( IconifyOp, KWM::isIconified( win ) );
    popup->setItemChecked( MaximizeOp, KWM::isMaximized( win ) );
}

void TaskButton::desktopPopupAboutToShow()
{
    if ( !desk_popup )
        return;
    desk_popup->clear();
    desk_popup->insertItem( i18n("&All desktops"), 0 );
    desk_popup->insertSeparator();
    if ( KWM::isSticky(win) )
        desk_popup->setItemChecked( 0, true );
    int id;
    for ( int i = 1; i <= KWin::numberOfDesktops(); i++ ) {
        id = desk_popup->insertItem( QString("&")+QString::number(i ), i );
        if ( !KWM::isSticky(win) && KWM::currentDesktop() == i )
            desk_popup->setItemChecked( id, TRUE );
    }
}

void TaskButton::clientPopupActivated( int id )
{
    switch ( id )
    {
    case MaximizeOp:
        KWM::setMaximize( win, !KWM::isMaximized( win ) );
        break;
    case IconifyOp:
        KWM::setIconify( win, !KWM::isIconified( win ) );
        break;
    case CloseOp:
        KWM::close( win );
        break;
    default:
        break;
    }
}

void TaskButton::sendToDesktop( int desk )
{
    if ( desk == 0 )
    {
        KWM::setSticky( win, !KWM::isSticky( win ) );
        return;
    }

    if ( KWM::isSticky( win ) )
        KWM::setSticky( win, false );

    if ( desk == KWin::currentDesktop() )
        return;

    KWM::moveToDesktop( win, desk );
}

void TaskButton::mouseReleaseEvent(QMouseEvent *e)
{
    if ( e->button() == QMouseEvent::RightButton )
    {
	    if (!popup)
        {
            popup = new QPopupMenu;
            popup->setCheckable( TRUE );
            connect( popup, SIGNAL( aboutToShow() ),
                    this,  SLOT( clientPopupAboutToShow() ) );
            connect( popup, SIGNAL( activated(int) ),
                    this,  SLOT( clientPopupActivated(int) ) );

            if (!desk_popup) desk_popup = new QPopupMenu( popup );
            desk_popup->setCheckable( TRUE );
            connect( desk_popup, SIGNAL( aboutToShow() ),
                    this,       SLOT( desktopPopupAboutToShow() ) );
            connect( desk_popup, SIGNAL( activated(int) ),
                    this,       SLOT( sendToDesktop(int) ) );

            popup->insertItem( i18n("Mi&nimize"), IconifyOp );
            popup->insertItem( i18n("Ma&ximize"), MaximizeOp );

            popup->insertSeparator();

            popup->insertItem(i18n("&To desktop"), desk_popup );

            popup->insertSeparator();

            popup->insertItem(i18n("&Close"), CloseOp);
        }
        popup->popup(e->globalPos());
    }
    QPushButton::mouseReleaseEvent(e);
}

QSizePolicy TaskButton::sizePolicy() const
{
    return(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred));
}

QSize TaskButton::sizeHint() const
{
    return(minimumSize());
}

