#include <qxembed.h>
#include <qtimer.h>
#include <qstringlist.h>
#include <qdragobject.h>
#include <qobjectlist.h>
#include <qpixmap.h>
#include <kglobal.h>
#include <kconfig.h>
#include <kstddirs.h>
#include <kiconloader.h>
#include <ksimpleconfig.h>
#include <kurl.h>
#include <kdesktopfile.h>
#include <kmimetype.h>
#include <dcopclient.h>
#include "itemarea.h"
#include "panelbutton.h"
#include "dialogs.h"

template class QList<PanelButton>;

ItemArea::ItemArea(const QString &configGroup, Orientation orientation,
                   int strut, QWidget* parent, const char* name)
  : Panner( orientation, strut, parent, name )
{
    btnSize = strut;
    configStr = configGroup;

    KConfig* config = KGlobal::config();
    config->setGroup("theme");
    QString bgStr = config->readEntry("itemBackground", "");
    if(!bgStr.isEmpty()){
        QPixmap bgPix(BarIcon(bgStr));
        if(!bgPix.isNull()){
            QBrush bgBrush(colorGroup().background(), bgPix);
            QPalette pal = kapp->palette();
            pal.setBrush(QColorGroup::Background, bgBrush);
            setPalette(pal);
        }
        else{
            warning("Kicker: Error loading itemArea theme pixmap");
        }
    }

    config->setGroup(configStr);
    QStringList buttonDataList;
    bool hasBtnConfig = config->hasKey("data");
    if(hasBtnConfig)
        buttonDataList = config->readListEntry("data");
    else{
        QString appPath = KGlobal::dirs()->resourceDirs("apps").last();
        buttonDataList.append("System");
        buttonDataList.append(" ");
        if(QFile::exists(appPath + "/Home.desktop")){
            buttonDataList.append("URL");
            buttonDataList.append(appPath + "/Home.desktop");
        }
        if(QFile::exists(appPath + "/Kfind.desktop")){
            buttonDataList.append("URL");
            buttonDataList.append(appPath + "/Kfind.desktop");
        }
        if(QFile::exists(appPath + "/KControl.desktop")){
            buttonDataList.append("URL");
            buttonDataList.append(appPath + "/KControl.desktop");
        }
        if(QFile::exists(appPath + "/System/konsole.desktop")){
            buttonDataList.append("URL");
            buttonDataList.append(appPath + "/System/konsole.desktop");
        }
    }

    QStringList::Iterator it;
    it = buttonDataList.begin();
    while(it != buttonDataList.end()) {
        PanelButton *btn = 0;
        QString typeStr(*it);
        ++it;
        if(typeStr == "System")
            btn = new PanelKButton(viewport());
        else if(typeStr == "Browser")
            btn = new PanelBrowserButton(*it, viewport());
        else if(typeStr == "URL")
            btn = new PanelURLButton(*it, viewport());
        else if(typeStr == "Exe")
            btn = new PanelExeButton(*it, viewport());
	else continue;
        
        if(btn) {
            btnList.append(btn);
            btn->show();
            connect(btn, SIGNAL(remove(PanelButton *)), this,
                    SLOT(removeButton(PanelButton *)));
            connect(btn, SIGNAL(requestSave()), this,
                    SLOT(saveButtonConfig()));
        }
        ++it;
    }
    layoutChildren();
    if(!hasBtnConfig)
        saveButtonConfig();
    setAcceptDrops(true);
}

void ItemArea::saveButtonConfig()
{
    QListIterator<PanelButton> it(btnList);
    KConfig *config = KGlobal::config();
    QStringList dataList;
    QString data;
    config->setGroup(configStr);
    for(; it.current(); ++it){
        switch(it.current()->type()){
        case PanelButton::SystemButton:
            dataList.append("System");
            break;
        case PanelButton::BrowserButton:
            dataList.append("Browser");
            break;
        case PanelButton::URLButton:
            dataList.append("URL");
            break;
        case PanelButton::UnknownExeButton:
            dataList.append("Exe");
            break;
        }
        it.current()->writeConfigData(data);
        dataList.append(data);
    }
    config->writeEntry("data", dataList);
    config->sync();
}


void ItemArea::setOrientation(Orientation orientation)
{
    Panner::setOrientation( orientation );
    //QListIterator<PanelButton> it(btnList);
    //for(; it.current(); ++it)
    //    it.current()->panelSizeChanged();
    layoutChildren();
}

ItemArea::~ItemArea()
{
}

void ItemArea::dragEnterEvent(QDragEnterEvent *ev)
{
    ev->accept(QUriDrag::canDecode(ev));
}

void ItemArea::dropEvent(QDropEvent *ev)
{
    QStrList fileList;
    if(QUriDrag::decode(ev, fileList)){
        QStrListIterator it(fileList);
        for(;it.current(); ++it){
            warning("In dropEvent for %s", it.current());
            PanelButton *btn;
            KURL url(it.current());
            // see if it's a executable or directory
            if(url.isLocalFile() && !KDesktopFile::isDesktopFile(url.path())){
                QFileInfo fi(url.path());
                if(fi.isDir()){ // directory
                    PanelDirDropMenu mnu;
                    switch(mnu.exec(mapToGlobal(ev->pos()))){
                    case PanelDirDropMenu::Browser:
                        btn = new PanelBrowserButton(url.path(), viewport());
                        break;
                    case PanelDirDropMenu::Url:
                    default:
                        btn = new PanelURLButton(it.current(), viewport());
                        break;
                    }
                }
                else if(fi.isExecutable()){ // non-KDE executable
                    warning("Dropped exe");
                    QString pixmapFile;
                    KMimeType::pixmapForURL(url.path(), 0,
                                            KIconLoader::Medium /*HACK*/, &pixmapFile);
                    PanelExeDialog dlg(url.path(), pixmapFile,
                                       QString::null, false, viewport());
                    if(dlg.exec() == QDialog::Accepted){
                        // KIconloader returns a full path, we only want name
                        QFileInfo iconfi(dlg.icon());
                        btn = new PanelExeButton(url.path(), iconfi.fileName(),
                                                 dlg.commandLine(),
                                                 dlg.useTerminal(), viewport());
                    }
                    else
                        break;
                }

                else // some unknown local file
                    btn = new PanelURLButton(it.current(), viewport());
            }
            else // a internet URL or desktop file.
                btn = new PanelURLButton(it.current(), viewport());
            btn->show();
            btnList.append(btn);
            connect(btn, SIGNAL(remove(PanelButton *)), this,
                    SLOT(removeButton(PanelButton *)));
            connect(btn, SIGNAL(requestSave()), this,
                    SLOT(saveButtonConfig()));
            saveButtonConfig();
        }
    }
    layoutChildren();
}

void ItemArea::addButton(PanelButton *btn)
{
    btn->show();
    btnList.append(btn);
    saveButtonConfig();
    layoutChildren();
    connect(btn, SIGNAL(repositioned()), this, SLOT(layoutChildren()));
    connect(btn, SIGNAL(remove(PanelButton *)), this,
            SLOT(removeButton(PanelButton *)));
    connect(btn, SIGNAL(requestSave()), this,
            SLOT(saveButtonConfig()));
}

void ItemArea::removeButton(PanelButton *btn)
{
    btn->hide();
    btnList.removeRef(btn);
    saveButtonConfig();
    layoutChildren();
}

void ItemArea::layoutChildren()
{
    int pos = 0;

    PanelButton *btn;
    QListIterator<PanelButton> it(btnList);
    for(; it.current() ; ++it) {
        btn = (PanelButton*)it.current();
        btn->setOrientation(orientation());
        btn->setPosition(position());
        btn->resize(btnSize, btnSize);
        btn->panelSizeChanged();
        if ( orientation() == Horizontal ) {
            btn->move( pos, 0 );
            pos += btn->width() + 1;
        }
        else { // Vertical
            btn->move( 0, pos );
            pos += btn->height() + 1;
        }
    }
    resizeContents(viewport()->childrenRect().width(),
                   viewport()->childrenRect().height());
}


