/*
 * (C) Daniel M. Duley (mosfet@kde.org) KDE artistic license.
 */

#include "menus.h"
#include "panel.h"
#include "panelbutton.h"
#include "itemarea.h"
#include <qdir.h>
#include <qdragobject.h>
#include <ksimpleconfig.h>
#include <kapp.h>
#include <klocale.h>
#include <kglobal.h>
#include <kstddirs.h>
#include <kiconloader.h>
#include <kdesktopfile.h>
#include <kmimetype.h>
#include <kurl.h>
#include <krun.h>
#include <kuserpaths.h>
#include <kwm.h>
#include <kwin.h>
#include <dcopclient.h>
#include <stdlib.h>
#include <unistd.h>

template class QList<PanelBrowserMenu>;

extern Panel *panel;

PanelMenu::PanelMenu(const QString &startDir, QWidget *parent, const char *name)
    : KPopupMenu(parent, name)
{
    init = false;
    startPath = startDir;
    connect(this, SIGNAL(activated(int)), SLOT(slotExec(int)));
    connect(this, SIGNAL(aboutToShow()), SLOT(slotAboutToShow()));
    KConfig *config = KGlobal::config();
    config->setGroup("menus");
    clearDelay = config->readNumEntry("MenuCacheTime", 60000); // 1 minute
    if(clearDelay)
        connect(&t, SIGNAL(timeout()), this, SLOT(slotClear()));
}

void PanelMenu::slotAboutToShow()
{
    if(clearDelay)
        t.stop();
    initialize();
}

void PanelMenu::slotClear()
{
    clear();
    init = false;
}

void PanelMenu::hideEvent(QHideEvent *ev)
{
    if(clearDelay)
        t.start(clearDelay, true);
    QPopupMenu::hideEvent(ev);
}

PanelBrowserMenu::PanelBrowserMenu(QString startDir, bool expandFSTypes,
                                   bool ignoreEmptyDirs, QWidget *parent,
                                   const char *name, int filter,
                                   int sorting)
    : PanelMenu(startDir, parent, name)
{
  expandFS = expandFSTypes;
  ignoreEmpty = ignoreEmptyDirs;
  KConfig *config = KGlobal::config();
  config->setGroup("menus");
  max = config->readNumEntry("MaxEntries", 100);
  merge = config->readBoolEntry("MergeKDEDirs", false);
  maxedOut=false;
  // gcc didn't like me using QDir:: for the type in the declaration
  filterSpec = (QDir::FilterSpec)filter;
  sortSpec = (QDir::SortSpec)sorting;
}

void PanelBrowserMenu::initialize(bool reInit)
{
    if (!init || reInit) {
        if (!reInit)
	  init = true;
        QDir dir(startPath);
        if(dir.count() > max){
            maxedOut = true;
            insertItem(KGlobal::iconLoader()->
                       loadIcon("exclamation", KIconLoader::Small),
                       i18n("Too many files to load."));
            return;
        }
        dir.setSorting(sortSpec);
        dir.setFilter(filterSpec);
        const QFileInfoList *list = dir.entryInfoList();
        if(!list){
            insertItem(KGlobal::iconLoader()->
                       loadIcon("stop", KIconLoader::Small),
                       i18n("Permission Denied!"));
            maxedOut = true;
            return;
        }
        QFileInfoListIterator it(*list);
        QFileInfo *fi;
        QPixmap smallIcon, medIcon;
        // just assume we will have a lot of these
        QIconSet defaultIcons(KGlobal::iconLoader()->
			      loadIcon("unknown", 
						  KIconLoader::Small));
	defaultIcons.setPixmap(KGlobal::iconLoader()->
			       loadIcon("unknown",
						   KIconLoader::Medium),
			       QIconSet::Large);

        QIconSet defaultFolders(KGlobal::iconLoader()->
				loadIcon("folder",
						    KIconLoader::Small));
	defaultFolders.setPixmap(KGlobal::iconLoader()->
				 loadIcon("folder",
						     KIconLoader::Medium),
				 QIconSet::Large);

        for(; (fi = it.current()) != NULL; ++it){
            // directories
            if(fi->isDir() && fi->fileName() != "." && fi->fileName() != ".."){
                if(ignoreEmpty){
                    QDir tmpDir(fi->absFilePath());
                    if((int)tmpDir.count() ==
                       (QFile::exists(fi->absFilePath()+"/.directory") ?
                        2 : 1)){
                        warning("Ignoring empty directory %s",
                                fi->fileName().latin1());
                        continue;
                    }
                }

		PanelBrowserMenu *subMenu = 0;
		if (merge && reInit) {
		  for (subMenu = subMenus.first(); subMenu != 0;
		       subMenu = subMenus.next()) {
		    if (QDir(subMenu->path()).dirName() == 
			fi->fileName())
		      break;
		  }
		}

		// if we didn't find the menu already (or we are not
		// merging menus) we need to create a new submenu.
		if (!subMenu) {
		  subMenu =
                    new PanelBrowserMenu(fi->absFilePath(), expandFS,
                                         ignoreEmpty, this);
		  subMenus.append(subMenu);
		  
		  QString dirFile(fi->absFilePath()+"/.directory");
		  QString entryName;
		  if(QFile::exists(dirFile)){
                    KSimpleConfig config(dirFile);
                    config.setDesktopGroup();
                    smallIcon = KGlobal::iconLoader()->
		      loadIcon(config.readEntry("Icon", "folder"), KIconLoader::Small);
		    medIcon = KGlobal::iconLoader()->
		      loadIcon(config.readEntry("Icon", "folder"), KIconLoader::Medium);
                    entryName = config.readEntry("Name", fi->fileName());
		  }
		  if (entryName.isNull())
		    entryName = fi->fileName();
		  if(smallIcon.isNull())
                    append(defaultFolders, entryName, subMenu);
		  else
                    append(smallIcon, medIcon, entryName, subMenu);
		} else {
		  QString oldPath = subMenu->path();
		  subMenu->setPath(fi->absFilePath());
		  subMenu->initialize(true);
		  subMenu->setPath(oldPath);
		}
	    } else if(fi->isFile() && fi->fileName() != ".directory"){
                // kdelnks
                if(KDesktopFile::isDesktopFile(fi->absFilePath())){
                    KDesktopFile dFile(fi->absFilePath());
                    if(dFile.tryExec()){
		      smallIcon = KGlobal::iconLoader()->
			loadIcon(dFile.readIcon(),
					    KIconLoader::Small);
		      medIcon = KGlobal::iconLoader()->
			loadIcon(dFile.readIcon(), 
					    KIconLoader::Medium);
		      if(smallIcon.isNull())
			smallIcon = KMimeType::
			pixmapForURL(KURL(fi->absFilePath()), 0, 
				     KIconLoader::Small);
			medIcon = KMimeType::
			pixmapForURL(KURL(fi->absFilePath()), 0,
				     KIconLoader::Medium);
                        if((dFile.hasDeviceType() || dFile.hasLinkType()) &&
                           expandFS){
                            // expand FS device applnks
                            KURL url(dFile.readURL());
                            PanelBrowserMenu *mnu =
                                new PanelBrowserMenu(url.path(), expandFS,
                                                     ignoreEmpty, this);
                            append(smallIcon, medIcon, fi->baseName(), mnu);
                        }
                        else{
                            if(!dFile.readName().isNull())
                                append(smallIcon, medIcon,
				       dFile.readName(), fi->fileName());
                            else{
                                QString nameStr = fi->baseName();
                                nameStr[0] = nameStr[0].upper();
                                append(smallIcon, medIcon,
				       nameStr, fi->fileName());
                            }
                        }
                    }
                }
                // normal files
                else{
                    QPixmap cachePix =
		      KMimeType::pixmapForURL(KURL(fi->absFilePath()), 0,
                                              KIconLoader::Small);
                    if(!cachePix.isNull())
                        append(cachePix, fi->fileName(), fi->fileName());
                    else
                        append(defaultIcons, fi->fileName(), fi->fileName());
                }
            }
            smallIcon.resize(0, 0);
	    medIcon.resize(0, 0);
        }
        adjustSize();
    }
}


void PanelBrowserMenu::append(const QPixmap &pixmap, const QString &title,
                           const QString &fileName)
{
    fileList.append(fileName);
    insertItem(QIconSet(pixmap), title, fileList.count()-1);
}

void PanelBrowserMenu::append(const QPixmap &smallPixmap, 
			      const QPixmap &mediumPixmap,
			      const QString &title,
			      const QString &fileName)
{
    fileList.append(fileName);
    QIconSet is(smallPixmap);
    is.setPixmap(mediumPixmap, QIconSet::Large);
    insertItem(is, title, fileList.count()-1);
}

void PanelBrowserMenu::append(const QIconSet &iconSet, const QString &title,
			      const QString &fileName)
{
    fileList.append(fileName);
    insertItem(iconSet, title, fileList.count()-1);
}

void PanelBrowserMenu::append(const QPixmap &pixmap, const QString &title,
			      PanelBrowserMenu *subMenu)
{
    fileList.append(""); // keep in sync
    insertItem(QIconSet(pixmap), title, subMenu, fileList.count()-1);
}

void PanelBrowserMenu::append(const QPixmap &smallPixmap, 
			      const QPixmap &mediumPixmap,
			      const QString &title,
			      PanelBrowserMenu *subMenu)
{
    QIconSet is(smallPixmap);
    is.setPixmap(mediumPixmap, QIconSet::Large);
    insertItem(is, title, subMenu, fileList.count()-1);
}

void PanelBrowserMenu::append(const QIconSet &iconSet, const QString &title,
			      PanelBrowserMenu *subMenu)
{
    insertItem(iconSet, title, subMenu, fileList.count()-1);
}

void PanelBrowserMenu::mousePressEvent(QMouseEvent *me)
{
  QPopupMenu::mousePressEvent(me);

  m_pressedPos = me->pos();
}

void PanelBrowserMenu::mouseMoveEvent(QMouseEvent *me)
{
  QPopupMenu::mouseMoveEvent(me);

  if (!(me->state() & LeftButton))
    return; // only handle left button drags

  int x = me->pos().x();
  int y = me->pos().y();
  
  if (abs(x - m_pressedPos.x()) > KGlobal::dndEventDelay() ||
      abs(y - m_pressedPos.y()) > KGlobal::dndEventDelay()) {
    int id = idAt(m_pressedPos);
    // the >=0 thing is a neat hack to avoid checking manual items
    if(id >= 0 && !maxedOut && !fileList[id].isEmpty()) {
      //      QUriDrag *d = new QUriDrag(findItem(id));
      QUriDrag *d = new QUriDrag(this);
      QPoint hotspot;
      hotspot.setX(iconSet(id)->pixmap(QIconSet::Large, 
				       QIconSet::Active).width() / 2);
      hotspot.setY(iconSet(id)->pixmap(QIconSet::Large, 
				       QIconSet::Active).height() / 2);
      d->setPixmap(iconSet(id)->pixmap(QIconSet::Large, 
				       QIconSet::Active), hotspot);
      
      QStringList l;
      l.append(startPath + "/" + fileList[id]);
      d->setFilenames(l);

      d->drag();
    }
  }
}

void PanelBrowserMenu::slotExec(int id)
{
    // the >=0 thing is a neat hack to avoid checking manual items
    if(id >= 0 && !maxedOut && !fileList[id].isEmpty())
        new KRun(startPath + "/" + fileList[id], 0, true);
}

PanelQuickBrowser::PanelQuickBrowser(QWidget *parent, const char *name)
    : PanelMenu("", parent, name)
{
    KConfig *config = KGlobal::config();
    config->setGroup("menus");
    max = config->readNumEntry("MaxEntries", 100);
}

void PanelQuickBrowser::initialize(bool)
{
    if(!init){
        init = true;
        // special handling for the device menu
        QDir dir( KUserPaths::desktopPath() );
        dir.setSorting(QDir::DirsFirst| QDir::Name | QDir::IgnoreCase);
        dir.setFilter(QDir::Files);
        const QFileInfoList *list = dir.entryInfoList();
        bool flag = false;
        QFileInfoListIterator it(*list);
        if(list){
          QFileInfo *fi;
          for(; (fi = it.current()) != NULL; ++it){
            KDesktopFile dFile(fi->absFilePath());
            if(dFile.hasDeviceType()){
                if(!flag){
                    insertTitle(KGlobal::iconLoader()->
                                loadIcon("blockdevice",
                                                    KIconLoader::Small),
                                i18n("Devices:"));
                    fileList.append("");
                    flag = true;
                }
                KURL url(dFile.readURL());
                PanelBrowserMenu *mnu =
                    new PanelBrowserMenu(url.path(), true, false, this);
                insertItem(KGlobal::iconLoader()->
                           loadIcon(dFile.readIcon(), KIconLoader::Small),
                           fi->baseName(), mnu);
                fileList.append("");
            }
          }
        }

        // use kdisknav's kdelnks for now
        int count = KGlobal::dirs()->findDirs("data", "quickbrowser").count();
        if(count){
            dir.setPath(KGlobal::dirs()->findDirs("data", "quickbrowser")
                        [count > 1 ? 1 : 0]);
            list = dir.entryInfoList();
            if(list){
                it = QFileInfoListIterator(*list);
                parseBrowserApplnks(it, "System Entries:",
                                    locate("mini", "shared.png"));
            }
            if(count > 1){
                dir.setPath(KGlobal::dirs()->findDirs("data", "quickbrowser")[0]);
                if(dir.exists()){
                    list = dir.entryInfoList();
                    if(list){
                        it = QFileInfoListIterator(*list);
                        parseBrowserApplnks(it, "Personal Entries:",
                                            locate("mini", "personal.png"));
                    }
                }
            }
        }
    }
}

void PanelQuickBrowser::parseBrowserApplnks(QFileInfoListIterator &it,
                                            const QString &title,
                                            const QString &pixmap)
{
    bool flag = false;
    QFileInfo *fi;
    for(; (fi = it.current()) != NULL; ++it){
        if(KDesktopFile::isDesktopFile(fi->absFilePath())){
            KDesktopFile dFile(fi->absFilePath());
            if(!flag){
                insertTitle(QPixmap(pixmap), title);
                fileList.append("");
                flag = true;
            }
            if(dFile.hasDeviceType() || dFile.hasLinkType()){
                KURL url(dFile.readURL());
                PanelBrowserMenu *mnu =
                    new PanelBrowserMenu(url.path(), true, false, this);
                insertItem(KGlobal::iconLoader()->
                           loadIcon(dFile.readIcon(), KIconLoader::Small),
                           fi->baseName(), mnu);
                fileList.append("");
            }
            else{
                insertItem(KGlobal::iconLoader()->
                           loadIcon(dFile.readIcon(), KIconLoader::Small),
                           fi->baseName());
                fileList.append(fi->absFilePath());
            }
        }
    }
}

void PanelQuickBrowser::slotExec(int id)
{
    if(!fileList[id].isEmpty() && id)
        new KRun(fileList[id]); // will delete itself
}

PanelKMenu::PanelKMenu(QWidget *parent, const char *name)
    : PanelBrowserMenu(KGlobal::dirs()->resourceDirs("apps").last(), false,
                       true, parent, name)
{

}

void PanelKMenu::initialize(bool)
{
    if(!init){
        PanelBrowserMenu::initialize();
        KIconLoader *ldr = KGlobal::iconLoader();
        QPopupMenu *sizeMnu = new QPopupMenu;
        sizeMnu->insertItem(i18n("Tiny"), this, SLOT(slotSizeSmall()));
        sizeMnu->insertItem(i18n("Normal"), this, SLOT(slotSizeNormal()));
        sizeMnu->insertItem(i18n("Large"), this, SLOT(slotSizeLarge()));

        QPopupMenu *panelOpMnu = new QPopupMenu;
        panelOpMnu->insertItem(i18n("Panel Size"), sizeMnu);
        panelOpMnu->insertItem(i18n("Add Application"));
        panelOpMnu->insertItem(i18n("Configure..."), this, SLOT(slotConfigure()));
        panelOpMnu->insertItem(i18n("Restart"), this, SLOT(slotRestart()));

        // testing
        KConfig *config = KGlobal::config();
        config->setGroup("menus");
	bool merge = config->readBoolEntry("MergeKDEDirs", false);
	QStringList appDirs =
	  KGlobal::dirs()->resourceDirs("apps");
	
	QStringList::Iterator it(appDirs.begin());
	for (; it != appDirs.end(); ++it) {
	  // we want to skip the last one, it's already done.
	  if (*it == appDirs.last())
  	    break;
	  QString dirFile(*it+"/.directory");
	  QString entryName;
	  QIconSet smallIcon;
	  if(QFile::exists(dirFile)){
	    KSimpleConfig config(dirFile);
	    config.setDesktopGroup();
	    smallIcon = KGlobal::iconLoader()->
	      loadIcon(config.readEntry("Icon", "folder"), KIconLoader::Small);
	    entryName = config.readEntry("Name", QDir(*it).dirName());
	  }

	  if (smallIcon.isNull()) {
	    QIconSet def(KGlobal::iconLoader()->
			 loadIcon("folder",
				  KIconLoader::Small));
	    def.setPixmap(KGlobal::iconLoader()->
			  loadIcon("folder",
				   KIconLoader::Medium),
			  QIconSet::Large);
	    smallIcon = def;
	  }
	  
	  if (merge) {
	    setPath(*it);
	    PanelBrowserMenu::initialize(true);
	  } else {
	    insertSeparator();
	    PanelBrowserMenu *newMnu =
	      new PanelBrowserMenu(*it, true, false, this);
	    insertItem(smallIcon, entryName, newMnu);
	  }
	}

	insertSeparator();

        bool flag = config->readBoolEntry("UseRecent", true);
        if(flag){
        	PanelRecentMenu *recentMnu = new PanelRecentMenu(this);
                insertItem(ldr->loadIcon("document", KIconLoader::Small),
                           i18n("Recent Documents"), recentMnu);
        	insertSeparator();
        }
        flag = config->readBoolEntry("UseBrowser", true);
        if(flag){
            PanelQuickBrowser *browserMnu = new PanelQuickBrowser(this);
            insertItem(ldr->loadIcon("folder_red", KIconLoader::Small),
                       i18n("Quick Browser"), browserMnu);
            insertSeparator();
        }
        // testing
        insertItem(ldr->loadIcon("panel", KIconLoader::Small),
		   i18n("Panel Menu"), panelOpMnu);
        insertItem( i18n("Context Help"), this, SLOT( slotContextHelp() ) );
        insertItem(ldr->loadIcon("key", KIconLoader::Small),
                   i18n("Lock Screen"), this, SLOT(slotLock()));
        insertItem(ldr->loadIcon("exit", KIconLoader::Small),
                   i18n("Logout"), this, SLOT(slotLogout()));
        adjustSize();
    }
}

void PanelKMenu::removeChildMenu(QPopupMenu *popup)
{
    int index;
    findPopup(popup, &index);
    if(index != -1){
        removeItemAt(index);
        adjustSize();
    }
}

void PanelKMenu::slotLock()
{
    DCOPClient *client = kapp->dcopClient();
    client->send("kdesktop", "KScreensaverIface", "lock()", "");
}

void PanelKMenu::slotLogout()
{
    KWM::logout();
}

void PanelKMenu::slotSizeSmall()
{
    panel->setUserSize(Panel::Tiny);
}

void PanelKMenu::slotSizeNormal()
{
    panel->setUserSize(Panel::Normal);
}

void PanelKMenu::slotSizeLarge()
{
    panel->setUserSize(Panel::Large);
}

void PanelKMenu::slotConfigure()
{
  
  // call kcmkpanel, or kcmkicker ?
}


void PanelKMenu::slotContextHelp()
{
    KWin::invokeContextHelp();
}

void PanelKMenu::slotRestart()
{
    char ** o_argv = new char*[2];
    // TODO : support for cmd line arguments
    //for (int v=0; v<o_argc; v++) o_argv[v] = argv[v];
    o_argv[0] = strdup("kicker");
    o_argv[1] = 0L;

    QApplication::exit();
    execvp(o_argv[0], o_argv);
    exit(1);
}

PanelRecentMenu::PanelRecentMenu(QWidget *parent, const char *name)
:PanelBrowserMenu(KGlobal::dirs()->resourceDirs("data").first() +
                  "/RecentDocuments/", false, false, parent, name,
                  QDir::Files, QDir::Time || QDir::Reversed)
{
    QString path(KGlobal::dirs()->resourceDirs("data").first() +
                 "/RecentDocuments/");
    if(!QFile::exists(path)){
        QDir d;
        d.mkdir(path);
    }
}

void PanelRecentMenu::initialize(bool)
{
    if(init)
        clear();
    init = false;
    PanelBrowserMenu::initialize();
    if(count()){
        insertSeparator();
        insertItem(KGlobal::iconLoader()->
                   loadIcon("delete", KIconLoader::Small),
                   i18n("Clear History"), this, SLOT(slotClearHistory())) ;
    }
}

void PanelRecentMenu::slotClearHistory()
{
    QDir dir(KGlobal::dirs()->resourceDirs("data").first() +
             "/RecentDocuments/");
    QStringList list = dir.entryList(QDir::Files);
    QStringList::Iterator it = list.begin();
    for(it = list.begin(); it != list.end() ; ++it)
        dir.remove(*it);
}



PanelButtonOpMenu::PanelButtonOpMenu(QWidget *parent, const char *name,
                                    bool removable)
    : KPopupMenu(parent, name)
{
    KIconLoader *ldr = KGlobal::iconLoader();
    // make a better icon for this
    insertTitle(ldr->loadIcon("go", KIconLoader::Small),
                i18n("Panel Item Options"));
    // Re-enable after Krash
    setItemEnabled(insertItem(BarIcon("move"), i18n("&Move"), Move), false);
    setAccel(CTRL+Key_M, Move);
    if(removable){
        insertItem(BarIcon("delete"),
                   i18n("&Remove"), Remove);
        setAccel(CTRL+Key_R, Remove);
    }
    insertSeparator();
    insertItem(ldr->loadIcon("kcontrol", KIconLoader::Small),
               i18n("&Properties"), Properties);
    setAccel(CTRL+Key_P, Properties);
    adjustSize();
}

PanelDirDropMenu::PanelDirDropMenu(QWidget *parent, const char *name)
    :KPopupMenu(parent, name)
{
    KIconLoader *ldr = KGlobal::iconLoader();
    insertTitle(ldr->loadIcon("folder", KIconLoader::Small),
                i18n("Panel Directory Options"));
    insertItem(i18n("Add as &file manager URL"), Url);
    setAccel(CTRL+Key_F, Url);
    insertItem(i18n("Add as Quick&Browser"), Browser);
    setAccel(CTRL+Key_B, Browser);
    adjustSize();
}


#include "menus.moc"

