#include "panelbutton.h"
#include "panel.h"
#include "itemarea.h"
#include "menus.h"
#include "dialogs.h"
#include <qtooltip.h>
#include <kglobal.h>
#include <kstddirs.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kpropsdlg.h>
#include <krun.h>
#include <kmessagebox.h>
#include <kdesktopfile.h>

PanelButton::PanelButton( QWidget * parent, const char *name)
    : QToolButton( parent, name )
{
    setBackgroundMode( PaletteBackground );
    setUsesBigPixmap( TRUE );
    setAutoRaise( TRUE );
    inMove = false;
    removeable = true;
}

void PanelButton::mousePressEvent(QMouseEvent *ev)
{
    if(inMove){
        releaseMouse();
        setCursor(arrowCursor);
        inMove = false;
        setMouseTracking(false);
    }
    else if(ev->button() == QMouseEvent::RightButton){
        PanelButtonOpMenu mnu(NULL, NULL, removeable);
        switch(mnu.exec(getPopupPosition(&mnu))){
        case PanelButtonOpMenu::Move:
            moveButton();
            break;
        case PanelButtonOpMenu::Remove:
            removeButton();
            break;
        case PanelButtonOpMenu::Properties:
            properties();
            break;
        default:
            break;
        }
    }
    else if(ev->button() == QMouseEvent::MidButton){
        grabMouse(sizeAllCursor);
        raise();
        inMove = true;
        mouseOffset = orientation() == Horizontal ? ev->x() : ev->y();
    }
    else
        QToolButton::mousePressEvent(ev);
}

void PanelButton::mouseMoveEvent(QMouseEvent *ev)
{
    if(inMove){
        if(orientation() == Horizontal)
            move(mapToParent(QPoint(ev->pos().x()-mouseOffset, y())));
        else
            move(mapToParent(QPoint(x(), ev->pos().y()-mouseOffset)));

    }
    else
        QToolButton::mouseMoveEvent(ev);
}

void PanelButton::mouseReleaseEvent(QMouseEvent *ev)
{
    if(ev->button() == QMouseEvent::MidButton){
        releaseMouse();
        setCursor(arrowCursor);
        inMove = false;
	emit repositioned();
    }
    else
        QToolButton::mouseReleaseEvent(ev);
}

void PanelButton::moveButton()
{
    inMove = true;
    raise();
    setMouseTracking(true);
    mouseOffset = width()/2;
    QCursor::setPos(mapToGlobal(QPoint(width()/2, width()/2)));
    grabMouse(sizeAllCursor);
}

void PanelButton::removeButton()
{
    emit remove(this);
    delete this;
    emit repositioned();
}

QPoint PanelButton::getPopupPosition(QPopupMenu *menu)
{
    if(position() == Bottom)
        return(mapToGlobal(QPoint(0, 0-menu->height())));
    else if(position() == Top)
        return(mapToGlobal(QPoint(0, height())));
    else if(position() == Right)
        return(mapToGlobal(QPoint(0-menu->width(), 0)));
    else
        return(mapToGlobal(QPoint(width(), 0)));
}

void PanelMenuButton::drawButtonLabel(QPainter *p)
{
    QToolButton::drawButtonLabel(p);
    if(width() > 32 && height() > 32){
        if(position() == Bottom)
            QApplication::style().drawArrow(p, Qt::UpArrow, isDown(), 3, 3,
                                            8, 8, colorGroup(), true);
        else if (position() == Top)
            QApplication::style().drawArrow(p, Qt::DownArrow, isDown(),
                                            3, height()-12, 8, 8,
                                            colorGroup(), true);
        else if (position() == Left)
            QApplication::style().drawArrow(p, Qt::RightArrow, isDown(),
                                            width()-12, 3, 8, 8,
                                            colorGroup(), true);
        else
            QApplication::style().drawArrow(p, Qt::LeftArrow, isDown(), 3, 3,
                                            8, 8, colorGroup(), true);
    }
}

PanelURLButton::PanelURLButton(const QString &url, QWidget *parent,
                               const char *name)
    : PanelButton(parent, name)
{
    urlStr = url;
    KURL u(url);
    local = u.isLocalFile();
    panelSizeChanged();
    connect(this, SIGNAL(clicked()), SLOT(slotExec()));
    setAcceptDrops(true);
}

void PanelURLButton::dragEnterEvent(QDragEnterEvent *ev)
{
    ev->accept(QUriDrag::canDecode(ev));
}

void PanelURLButton::dropEvent(QDropEvent *ev)
{
    QStrList fileList;
    QStringList execList;
    if(QUriDrag::decode(ev, fileList)){
        QStrListIterator it(fileList);
        for(;it.current(); ++it)
            execList.append(it.current());
        KURL url(urlStr);
        if(!execList.isEmpty() && KDesktopFile::isDesktopFile(url.path())){
            KService service(url.path());
            KRun::run(service, execList);
        }
    }
}

void PanelURLButton::panelSizeChanged()
{
    // width and height will be the same
    if(width() <  32 ){
        setUsesBigPixmap(false);
        setPixmap(KMimeType::pixmapForURL(KURL(urlStr), 0,
                                          KIconLoader::Small));
    }
    else if(width() < 48){
        setUsesBigPixmap(true);
        setPixmap(KMimeType::pixmapForURL(KURL(urlStr), 0,
                                          KIconLoader::Medium));
    }
    else{
        setUsesBigPixmap(true);
        setPixmap(KMimeType::pixmapForURL(KURL(urlStr), 0,
                                             KIconLoader::Large));
}
}

void PanelURLButton::slotExec()
{
    new KRun(urlStr, 0, local);
}

void PanelURLButton::properties()
{
    new PropertiesDialog(urlStr); // will delete itself
    /*
    setPixmap(QPixmap(KPixmapCache::pixmapFileForURL(urlStr.latin1(), 0,
                                                     local, false)));
    */
    // The setPixmap calls should be done out of panelSizeChanged (David)
    panelSizeChanged();
}

PanelBrowserButton::PanelBrowserButton(const QString &startDir, QWidget *parent,
                            const char *name)
    : PanelMenuButton(parent, name)
{
    topMenu = new PanelBrowserMenu(startDir);
    connect(this, SIGNAL(pressed()), SLOT(slotExecMenu()));
    panelSizeChanged();
}

void PanelBrowserButton::panelSizeChanged()
{
    QString dirFile(topMenu->path()+"/.directory");
    QString iconFile;
    if(QFile::exists(dirFile)){
        KSimpleConfig config(dirFile);
        config.setDesktopGroup();
        iconFile = config.readEntry("Icon", "folder");
    }
    if(iconFile.isEmpty())
        iconFile = "folder";

    if(width() < 32){
        setUsesBigPixmap(false);
        setPixmap(KGlobal::iconLoader()->
                  loadIcon(iconFile, KIconLoader::Small));
    }
    else if(width() < 48){
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->
                  loadIcon(iconFile));
    }
    else{
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->
                  loadIcon(iconFile, KIconLoader::Large));
    }
}

void PanelBrowserButton::slotExecMenu()
{
    topMenu->initialize();
    topMenu->popup(getPopupPosition(topMenu));
}


void PanelBrowserButton::properties()
{
    KMessageBox::sorry(this, i18n("Properties for QuickBrowser not implemented!"),
                       i18n("Apology from mosfet"));
}

PanelKButton::PanelKButton(QWidget *parent, const char *name)
    :PanelMenuButton(parent, name)
{
    panelSizeChanged();
    QToolTip::add(this, i18n("KDE Application Menu"));
    topMenu = new PanelKMenu();
    connect(this, SIGNAL(pressed()), SLOT(slotExecMenu()));
    removeable = false;
    // client menu stuff
    menuMgr = new KickerMenuManager( topMenu, this, "kickerMenuManager" );
    connect(menuMgr, SIGNAL(popupKMenu(int, int)), this,
                    SLOT(slotExecMenuAt(int, int)));
}

void PanelKButton::panelSizeChanged()
{
    if(width() < 32){
        setUsesBigPixmap(false);
        setPixmap(KGlobal::iconLoader()->loadIcon("go",
                                                             KIconLoader::Small));
    }
    else if(width() < 48){
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->loadIcon("go"));
    }
    else{
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->
                  loadIcon("go", KIconLoader::Large));
    }
}

void PanelKButton::slotExecMenu()
{
    topMenu->initialize();
    topMenu->popup(getPopupPosition(topMenu));
    topMenu->exec();
    setDown( FALSE );
}

void PanelKButton::slotExecMenuAt(int x, int y)
{
    topMenu->initialize();
    topMenu->popup(QPoint(x, y));
    topMenu->exec();
}

void PanelKButton::properties()
{
    KMessageBox::sorry(this, i18n("Properties for K Menu not implemented!"),
                       i18n("Apology from mosfet"));
}

// Non-KDE application
PanelExeButton::PanelExeButton(const QString &filePath, const QString &icon,
                               const QString &cmdLine, bool inTerm,
                               QWidget *parent, const char *name)
    : PanelButton(parent, name)
{
    pathStr = filePath;
    iconStr = icon;
    cmdStr = cmdLine;
    term = inTerm;

    connect(this, SIGNAL(clicked()), SLOT(slotExec()));
    panelSizeChanged();
}

PanelExeButton::PanelExeButton(const QString &configData, QWidget *parent,
                               const char *name)
    : PanelButton(parent, name)
{
    term = (configData[0] ==  '1') ? true : false;
    pathStr= configData.mid(1);
    int index = pathStr.find("::mossie::");
    pathStr.truncate(index);
    iconStr=configData.mid(index+11);
    int index2 = iconStr.find("::mossie::");
    iconStr.truncate(index2);
    cmdStr =configData.mid(index+11+index2+10);
    warning("term %d, path %s, icon %s, cmd %s", term, pathStr.latin1(),
            iconStr.latin1(), cmdStr.latin1());
    connect(this, SIGNAL(clicked()), SLOT(slotExec()));
    panelSizeChanged();
    setAcceptDrops(true);
}

void PanelExeButton::dragEnterEvent(QDragEnterEvent *ev)
{
    ev->accept(QUriDrag::canDecode(ev));
}

void PanelExeButton::dropEvent(QDropEvent *ev)
{
    QStrList fileList;
    QStringList blah;
    QString execStr;
    if(QUriDrag::decode(ev, fileList)){
        QStrListIterator it(fileList);
        for(;it.current(); ++it){
            KURL url(it.current());
            if(KDesktopFile::isDesktopFile(url.path())){
                KDesktopFile deskFile(url.path());
                deskFile.setDesktopGroup();
                execStr += deskFile.readURL() + " ";
            }
            else
                execStr += url.path() + " ";
        }
        bool result;
        if(term){
            KConfig *config = KGlobal::config();
            config->setGroup("misc");
            QString termStr = config->readEntry("Terminal", "konsole");
            result = KRun::run(termStr + " -e " + pathStr + " " +
                               cmdStr + " " + execStr, blah);

        }
        else
            result = KRun::run(pathStr + " " + cmdStr + " " + execStr, blah);

        if(!result)
            KMessageBox::error(this, i18n("Connect execute non-KDE application!"),
                               i18n("Kicker Error!"));
    }
}

void PanelExeButton::panelSizeChanged()
{
    if(width() < 32){
        setUsesBigPixmap(false);
        setPixmap(KGlobal::iconLoader()->loadIcon(iconStr, KIconLoader::Small));
        if(pixmap() == NULL)
	  setPixmap(KGlobal::iconLoader()->
		    loadIcon("exec", KIconLoader::Small));
        repaint(); // hack to update if icon was scaled (doesn't otherwise).
    }
    else if(width() < 48){
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->loadIcon(iconStr));
        if(pixmap() == NULL)
            setPixmap(KGlobal::iconLoader()->loadIcon("exec"));
    }
    else{
        setUsesBigPixmap(true);
        setPixmap(KGlobal::iconLoader()->loadIcon(iconStr));
        if(pixmap() == NULL)
            setPixmap(KGlobal::iconLoader()->
                      loadIcon("exec", KIconLoader::Large));
    }
}

void PanelExeButton::writeConfigData(QString &data)
{
    data.sprintf("%d%s::mossie::%s::mossie::%s", term, pathStr.latin1(),
                 iconStr.latin1(), cmdStr.latin1());
}

void PanelExeButton::slotExec()
{
    QStringList blah;
    bool result;
    if(term){
        KConfig *config = KGlobal::config();
        config->setGroup("misc");
        QString termStr = config->readEntry("Terminal", "konsole");
        result = KRun::run(termStr + " -e " + pathStr + " " + cmdStr, blah);
    }
    else
        result = KRun::run(pathStr + " " + cmdStr, blah);
    if(!result)
        KMessageBox::error(this, i18n("Connect execute non-KDE application!"),
                           i18n("Kicker Error!"));
}

void PanelExeButton::properties()
{
    PanelExeDialog dlg(pathStr, iconStr, cmdStr, term, this);
    if(dlg.exec() == QDialog::Accepted){
        // KIconloader returns a full path, we only want name
        QFileInfo iconfi(dlg.icon());
        iconStr = iconfi.baseName();
        cmdStr = dlg.commandLine();
        term = dlg.useTerminal();
        setPixmap(KGlobal::iconLoader()->loadIcon(iconStr));
        if(!pixmap())
            setPixmap(KGlobal::iconLoader()->loadIcon("exec"));
        emit requestSave();
    }
}

#include "panelbutton.moc"


