#ifndef PANELBUTTON_H
#define PANELBUTTON_H

#include "menus.h"
#include "panel.h"
#include <qtoolbutton.h>
#include <kurl.h>

class PanelButton: public QToolButton
{
    Q_OBJECT
public:
    enum ButtonType {SystemButton, URLButton, BrowserButton,
        UnknownExeButton};

    PanelButton(QWidget * parent, const char *name = 0);
    // used for menu popup
    void setPosition(Position p){pos = p;}
    void setOrientation(Orientation o){orient = o;}
    Position position(){return(pos);}
    Orientation orientation(){return(orient);}
    QPoint getPopupPosition(QPopupMenu *menu);
    // used for configuration
    virtual void writeConfigData(QString &data)=0;
    virtual void panelSizeChanged()=0; // don't use resizeEvent
    virtual ButtonType type()=0;
signals:
    void repositioned();
    void remove(PanelButton *btn);
    void requestSave();
protected:
    virtual void moveButton();
    virtual void removeButton();
    virtual void properties()=0;

    virtual void mousePressEvent(QMouseEvent *ev);
    virtual void mouseMoveEvent(QMouseEvent *ev);
    virtual void mouseReleaseEvent(QMouseEvent *ev);

    bool inMove, removeable;
    int mouseOffset;
    Position pos;
    Orientation orient;
};

/**
 * Base class for buttons with menus.
 */
class PanelMenuButton : public PanelButton
{
    Q_OBJECT
public:
    PanelMenuButton(QWidget *parent, const char *name)
        : PanelButton(parent, name){;}
protected:
    void drawButtonLabel(QPainter *p);
};

/**
 * Simple URL button (applnks, files, whatever)
 */
class PanelURLButton : public PanelButton
{
    Q_OBJECT
public:
    PanelURLButton(const QString &url, QWidget *parent=0,
                   const char *name=0);
    virtual ButtonType type(){return(URLButton);}
    virtual void writeConfigData(QString &data){data = urlStr;}
    virtual void panelSizeChanged();
protected slots:
    void slotExec();
protected:
    virtual void dropEvent(QDropEvent *ev);
    virtual void dragEnterEvent(QDragEnterEvent *ev);
    virtual void properties();
    QString urlStr;
    bool local;
};
    

/**
 * Button that contains a Browser directory menu
 */
class PanelBrowserButton : public PanelMenuButton
{
    Q_OBJECT
public:
    PanelBrowserButton(const QString &startDir, QWidget *parent=0,
                  const char *name=0);
    ~PanelBrowserButton(){;}
    virtual ButtonType type(){return(BrowserButton);}
    virtual void writeConfigData(QString &data){data = topMenu->path();}
    virtual void panelSizeChanged();
protected slots:
    void slotExecMenu();
protected:
    virtual void properties();

    PanelBrowserMenu *topMenu;
};

/**
 * Button that contains the PanelKMenu and client menu manager.
 */
class PanelKButton : public PanelMenuButton
{
    Q_OBJECT
public:
    PanelKButton(QWidget *parent=0, const char *name=0);
    virtual ButtonType type(){return(SystemButton);}
    virtual void writeConfigData(QString &data){data = " ";}
    virtual void panelSizeChanged();
protected slots:
    void slotExecMenu();
    void slotExecMenuAt(int x, int y);
protected:
    virtual void properties();
    PanelKMenu *topMenu;
    KickerMenuManager* menuMgr;
};

/**
 * Button that contains a non-KDE application
 */
class PanelExeButton : public PanelButton
{
    Q_OBJECT
public:
    PanelExeButton(const QString &filePath, const QString &icon,
                   const QString &cmdLine, bool inTerm, QWidget *parent=0,
                   const char *name=0);
    PanelExeButton(const QString &configData, QWidget *parent=0,
                   const char *name=0);
    virtual void writeConfigData(QString &data);
    virtual void panelSizeChanged();
    virtual ButtonType type(){return(UnknownExeButton);}
protected slots:
    void slotExec();
protected:
    virtual void dropEvent(QDropEvent *ev);
    virtual void dragEnterEvent(QDragEnterEvent *ev);
    virtual void properties();
    QString pathStr, iconStr, cmdStr;
    bool term;
};

#endif

